
#include <cstdarg>
#include <cstring>
#include <cstdio>
#include <cctype>
#include <vector>

#include "lg/interface.h"
#include "lg/scrservices.h"

#include "ScriptLib.h"


IMalloc* g_pMalloc;
IScriptMan* g_pScriptManager;

#ifdef _HACKS
bool g_fIsEditor;
char* g_pszFileName;
#endif

const cMultiParm g_mpUndef;


/* Whoops. msvcrt doesn't include strsep.
 */
char* strsep(char** str, const char* delim)
{
	register char* p = *str;
	if (!p) return NULL;
	while (*p)
	{
		register const char* d = delim;
		while (*d)
		{
			if (*p == *d++)
			{
				register char* t = *str;
				*p++ = '\0';
				*str = p;
				return t;
			}
		}
		++p;
	}
	p = *str;
	*str = NULL;
	return p;
}

/* Seperate a string on any of delim characters while preserving 
 * substrings surrounded by an equal pair of any of quotes characters.
 * str is the address of a pointer to the string to be seperated.
 * The pointer to the next token in the string, or NULL, is copied to 
 * the address and the the value that was passed in *str is returned.
 * delim and quotes cannot be NULL.
 * Empty tokens will be returned.
 * Like strtok and strsep, the string data will be modified.
 * str, of course, must be an lvalue.
 */
char* strqsep(char** str, const char* delim, const char* quotes)
{
	register char* p = *str;
	if (!p) return NULL;
	while (*p)
	{
		{
			register const char* q = quotes;
			while (*q)
			{
				if (*p == *q++)
				{
					register char* t = p;
				L_ESCAPEQUOTE:
					t = strchr(t+1, *p);
					if (!t)
					{
						p = *str;
						*str = NULL;
						return p;
					}
					if (*(t-1) == '\\')
						goto L_ESCAPEQUOTE;
					p = t;
					goto L_QUOTE;
				}
			}
		}
		{
			register const char* d = delim;
			while (*d)
			{
				if (*p == *d++)
				{
					register char* t = *str;
					*p++ = '\0';
					*str = p;
					return t;
				}
			}
		}
	L_QUOTE:
		++p;
	}
	p = *str;
	*str = NULL;
	return p;
}

char* GetObjectParams(int iObject)
{
	SService<IPropertySrv> pPS(g_pScriptManager);
	cMultiParm mpDesignNote;
	if (!pPS->Possessed(iObject, "DesignNote"))
		return NULL;
	pPS->Get(mpDesignNote, iObject, "DesignNote", NULL);
	return (mpDesignNote.psz);
}

int SetObjectParams(int iObject, const char* pszParams)
{
	SService<IPropertySrv> pPS(g_pScriptManager);
	cMultiParm mpDesignNote = const_cast<char*>(pszParams);
	return pPS->Set(iObject, "DesignNote", mpDesignNote);
}

char* GetObjectParamString(int iObject, const char* pszParam)
{
	char* pszString = GetObjectParams(iObject);
	char* pszRet = GetParamString(pszString, pszParam);
	g_pMalloc->Free(pszString);
	return pszRet;
}

int GetObjectParamInt(int iObject, const char* pszParam)
{
	char* params = GetObjectParams(iObject);
	if (!params)
		return 0;
	int iRet = GetParamInt(params, pszParam);
	g_pMalloc->Free(params);
	return iRet;
}

float GetObjectParamFloat(int iObject, const char* pszParam)
{
	char* params = GetObjectParams(iObject);
	if (!params)
		return 0;
	float fRet = GetParamFloat(params, pszParam);
	g_pMalloc->Free(params);
	return fRet;
}

char* GetParamString(const char* pszString, const char* pszParam)
{
	if (!pszString || !pszParam)
		return NULL;
	char* pszRet = NULL;
	char* pszTemp = new char[strlen(pszString) + 1];
	strcpy(pszTemp, pszString);
	char* pszSep, *pszToken;
	for (pszSep = pszTemp, pszToken = strqsep(&pszSep, ";", "\"'"); 
	     pszToken; pszToken = strqsep(&pszSep, ";", "\"'"))
	{
		while (isspace(*pszToken)) ++pszToken;
		char* pszStart = strchr(pszToken, '=');
		if (pszStart)
			*pszStart++ = '\0';
		if (!stricmp(pszToken, pszParam))
		{
			if (!pszStart)
				break;
			while (isspace(*pszStart)) ++pszStart;
			char* pszEnd;
			if (*pszStart == '\"')
			{
				++pszStart;
				pszEnd = strrchr(pszStart, '\"');
			}
			else if (*pszStart == '\'')
			{
				++pszStart;
				pszEnd = strrchr(pszStart, '\'');
			}
			else
				pszEnd = pszStart + strlen(pszStart);
			int iLen = pszEnd - pszStart;
			pszRet = reinterpret_cast<char*>(g_pMalloc->Alloc(iLen + 1));
			if (pszRet)
			{
				strncpy(pszRet, pszStart, iLen);
				pszRet[iLen] = '\0';
			}
			break;
		}
	}
	delete[] pszTemp;
	return pszRet;
}

int GetParamInt(const char* pszString, const char* pszParam)
{
	if (!pszString || !pszParam)
		return 0;
	int iRet = 0;
	char* pszTemp = new char[strlen(pszString) + 1];
	strcpy(pszTemp, pszString);
	char* pszSep, *pszToken;
	for (pszSep = pszTemp, pszToken = strqsep(&pszSep, ";", "\"'"); 
	     pszToken; pszToken = strqsep(&pszSep, ";", "\"'"))
	{
		while (isspace(*pszToken)) ++pszToken;
		char* pszStart = strchr(pszToken, '=');
		if (pszStart)
			*pszStart++ = '\0';
		if (!stricmp(pszToken, pszParam))
		{
			while (isspace(*pszStart)) ++pszStart;
			if (*pszStart == '$')
			{
				char* pszEnd = ++pszStart;
				while (isalnum(*pszEnd) || (*pszEnd == '_'))
					++pszEnd;
				*pszEnd = '\0';
				if (pszEnd > pszStart)
				{
					SService<IQuestSrv> pQS(g_pScriptManager);
					if (pQS->Exists(pszStart))
						iRet = pQS->Get(pszStart);
				}
			}
			else
			{
				if (*pszStart == '\"' || *pszStart == '\'')
					++pszStart;
				iRet = strtol(pszStart, NULL, 10);
			}
			break;
		}
	}
	delete[] pszTemp;
	return iRet;
}

float GetParamFloat(const char* pszString, const char* pszParam)
{
	if (!pszString || !pszParam)
		return 0;
	float fRet = 0;
	char* pszTemp = new char[strlen(pszString) + 1];
	strcpy(pszTemp, pszString);
	char* pszSep, *pszToken;
	for (pszSep = pszTemp, pszToken = strqsep(&pszSep, ";", "\"'"); 
	     pszToken; pszToken = strqsep(&pszSep, ";", "\"'"))
	{
		while (isspace(*pszToken)) ++pszToken;
		char* pszStart = strchr(pszToken, '=');
		if (pszStart)
			*pszStart++ = '\0';
		if (!stricmp(pszToken, pszParam))
		{
			while (isspace(*pszStart)) ++pszStart;
			if (*pszStart == '\"' || *pszStart == '\'')
				++pszStart;
			fRet = strtod(pszStart, NULL);
			break;
		}
	}
	delete[] pszTemp;
	return fRet;
}

int RemoveObjectParam(int iObject, const char* pszParam)
{
	char* pszString = GetObjectParams(iObject);
	if (!pszString)
		return 0;
	char* pszNew = new char[strlen(pszString) + 1];
	*pszNew = '\0';
	char* pszSep, *pszToken;
	for (pszSep = pszString, pszToken = strqsep(&pszSep, ";", "\"'"); 
	     pszToken; pszToken = strqsep(&pszSep, ";", "\"'"))
	{
		while (isspace(*pszToken)) ++pszToken;
		if (!*pszToken)
			continue;
		char* pszStart = strchr(pszToken, '=');
		if (pszStart)
			*pszStart++ = '\0';
		if (stricmp(pszToken, pszParam))
		{
			strcat(pszNew, pszToken);
			if (pszStart)
			{
				strcat(pszNew, "=");
				strcat(pszNew, pszStart);
			}
			strcat(pszNew, ";");
		}
	}
	SetObjectParams(iObject, pszNew);
	delete[] pszNew;
	g_pMalloc->Free(pszString);
	return 0;
}


static int DoAddMetaProperty(IObjectSrv* pOS, object oMetaProp, object oObject)
{
	int iRet = 0;
	true_bool has_mp;
	pOS->HasMetaProperty(has_mp, oObject, oMetaProp);
	if (!has_mp)
	{
		iRet = pOS->AddMetaProperty(oObject, oMetaProp);
	}
	return iRet;
}

static int DoRemoveMetaProperty(IObjectSrv* pOS, object oMetaProp, object oObject)
{
	int iRet = 0;
	true_bool has_mp;
	pOS->HasMetaProperty(has_mp, oObject, oMetaProp);
	if (has_mp)
	{
		iRet = pOS->RemoveMetaProperty(oObject, oMetaProp);
		//pOS->HasMetaProperty(has_mp, oObject, oMetaProp);
	}
	return iRet;
}

int AddSingleMetaProperty(const char* pszMetaProp, int iObjId)
{
	SService<IObjectSrv> pOS(g_pScriptManager);
	object oMetaProp;
	pOS->Named(oMetaProp, pszMetaProp);
	if (oMetaProp != 0)
		return DoAddMetaProperty(pOS, oMetaProp, iObjId);
	return 0;
}

int AddSingleMetaProperty(int iMetaProp, int iObjId)
{
	SService<IObjectSrv> pOS(g_pScriptManager);
	return DoAddMetaProperty(pOS, iMetaProp, iObjId);
}

int RemoveSingleMetaProperty(const char* pszMetaProp, int iObjId)
{
	SService<IObjectSrv> pOS(g_pScriptManager);
	object oMetaProp;
	pOS->Named(oMetaProp, pszMetaProp);
	if (oMetaProp != 0)
		return DoRemoveMetaProperty(pOS, oMetaProp, iObjId);
	return 0;
}

int RemoveSingleMetaProperty(int iMetaProp, int iObjId)
{
	SService<IObjectSrv> pOS(g_pScriptManager);
	return DoRemoveMetaProperty(pOS, iMetaProp, iObjId);
}


int GetOneLinkDest(const char* pszFlavor, int iSource, int* piRealSource)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	link l;
	pLS->GetOne(l, pLTS->LinkKindNamed(pszFlavor), iSource, 0);
	if (l == 0)
		return 0;
	sLink sl;
	pLTS->LinkGet(l, &sl);
	if (piRealSource)
		*piRealSource = sl.source;
	return sl.dest;
}

long GetAnyLink(const char* pszFlavor, int iSource, int iDest, sLink* pResult)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);

	std::vector<long> vLinks;
	linkset lsLinks;
	linkkind lkFlavor = (pszFlavor) ? pLTS->LinkKindNamed(pszFlavor) : 0;
	pLS->GetAll(lsLinks, lkFlavor, iSource, iDest);
	for (true_bool __p; lsLinks.AnyLinksLeft(__p), __p; lsLinks.NextLink())
	{
		vLinks.insert(vLinks.begin(), lsLinks.query->ID());
	}

	SService<IDataSrv> pDS(g_pScriptManager);
	long iLinkID = vLinks.at(pDS->RandInt(0, vLinks.size() - 1));
	if (pResult)
		pLTS->LinkGet(iLinkID, pResult);
	return iLinkID;
}

long FindLinkByData(const char* pszFlavor, int iSource, int iDest, sLink* pResult, const void* pvData, int iDataLen)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);

	int __cdecl (*cmp)(const void*, const void*, size_t) = memcmp;
	if (pvData && iDataLen == -1)
	{
		iDataLen = strlen(reinterpret_cast<const char*>(pvData)) + 1;
		cmp = reinterpret_cast<int (*)(const void*,const void*,size_t)>(strnicmp);
	}

	linkset lsLinks;
	linkkind lkFlavor = (pszFlavor) ? pLTS->LinkKindNamed(pszFlavor) : 0;
	pLS->GetAll(lsLinks, lkFlavor, iSource, iDest);
	long iLinkID = 0;
	for (true_bool __p; lsLinks.AnyLinksLeft(__p), __p; lsLinks.NextLink())
	{
		void* pLinkData = lsLinks.query->Data();
		if (pvData)
		{
			if (!pLinkData)
				continue;
			if ((*cmp)(pvData, pLinkData, iDataLen))
				continue;
		}
		else
		{
			if (pLinkData && *reinterpret_cast<const char*>(pLinkData))
				continue;
		}
		iLinkID = lsLinks.query->ID();
		if (pResult)
			lsLinks.query->Link(pResult);
	}
	return iLinkID;
}

long GetAnyLinkByData(const char* pszFlavor, int iSource, int iDest, sLink* pResult, const void* pvData, int iDataLen)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);

	int __cdecl (*cmp)(const void*, const void*, size_t) = memcmp;
	if (pvData && iDataLen == -1)
	{
		iDataLen = strlen(reinterpret_cast<const char*>(pvData)) + 1;
		cmp = reinterpret_cast<int (*)(const void*,const void*,size_t)>(strnicmp);
	}

	std::vector<long> vLinks;
	linkset lsLinks;
	linkkind lkFlavor = (pszFlavor) ? pLTS->LinkKindNamed(pszFlavor) : 0;
	pLS->GetAll(lsLinks, lkFlavor, iSource, iDest);
	for (true_bool __p; lsLinks.AnyLinksLeft(__p), __p; lsLinks.NextLink())
	{
		void* pLinkData = lsLinks.query->Data();
		if (pvData)
		{
			if (!pLinkData)
				continue;
			if ((*cmp)(pvData, pLinkData, iDataLen))
				continue;
		}
		else
		{
			if (pLinkData && *reinterpret_cast<const char*>(pLinkData))
				continue;
		}
		vLinks.insert(vLinks.begin(), lsLinks.query->ID());
	}

	SService<IDataSrv> pDS(g_pScriptManager);
	long iLinkID = vLinks.at(pDS->RandInt(0, vLinks.size() - 1));
	if (pResult)
		pLTS->LinkGet(iLinkID, pResult);
	return iLinkID;
}

long IterateLinks(const char* pszFlavor, int iSource, int iDest,
			    LinkIterCallbackFunc pfnLinkIterCallback, IScript* pScript, void* pData)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);

	linkset lsLinks;
	linkkind lkFlavor = (pszFlavor) ? pLTS->LinkKindNamed(pszFlavor) : 0;
	pLS->GetAll(lsLinks, lkFlavor, iSource, iDest);
	long iCount = 0;
	for (true_bool __p; lsLinks.AnyLinksLeft(__p), __p; lsLinks.NextLink())
	{
		++iCount;
		// It might be worthwhile to wrap this in a try .. catch
		if (!pfnLinkIterCallback(pLS, lsLinks.query, pScript, pData))
			break;
	}
	return iCount;
}

long IterateLinksByData(const char* pszFlavor, int iSource, int iDest,
			    const void* pvFilter, int iFilterLen,
			    LinkIterCallbackFunc pfnLinkIterCallback, IScript* pScript, void* pData)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);

	int __cdecl (*cmp)(const void*, const void*, size_t) = memcmp;
	if (pvFilter && iFilterLen == -1)
	{
		iFilterLen = strlen(reinterpret_cast<const char*>(pvFilter)) + 1;
		cmp = reinterpret_cast<int (*)(const void*,const void*,size_t)>(strnicmp);
	}

	linkset lsLinks;
	linkkind lkFlavor = (pszFlavor) ? pLTS->LinkKindNamed(pszFlavor) : 0;
	pLS->GetAll(lsLinks, lkFlavor, iSource, iDest);
	long iCount = 0;
	for (true_bool __p; lsLinks.AnyLinksLeft(__p), __p; lsLinks.NextLink())
	{
		void* pLinkData = lsLinks.query->Data();
		if (pvFilter)
		{
			if (!pLinkData)
				continue;
			if ((*cmp)(pvFilter, pLinkData, iFilterLen))
				continue;
		}
		else
		{
			if (pLinkData && *reinterpret_cast<const char*>(pLinkData))
				continue;
		}
		++iCount;
		// It might be worthwhile to wrap this in a try .. catch
		if (!pfnLinkIterCallback(pLS, lsLinks.query, pScript, pData))
			break;
	}
	return iCount;
}

static int FixupPlayerLinksFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	// Move the destination of the link to the object specified in pData
	SInterface<ILinkManager> pLM(g_pScriptManager);
	int iPlayer = reinterpret_cast<int>(pData);
	sLink slInfo;
	pLQ->Link(&slInfo);
	pLM->AddFull(slInfo.source, iPlayer, slInfo.flavor, pLQ->Data());
	pLM->Remove(pLQ->ID());
	return 1;
}

void FixupPlayerLinks(int iSource, int iPlayer)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	// Get the PlayerFactory link. There should only ever be one of these.
	link lPFLink;
	pLS->GetOne(lPFLink, pLTS->LinkKindNamed("PlayerFactory"), 0, 0);
	if (lPFLink == 0)
		return;
	// Iterate through all links from here to the source of the PF link.
	sLink lPFInfo;
	pLTS->LinkGet(lPFLink, &lPFInfo);
	IterateLinks(NULL, iSource, lPFInfo.source, FixupPlayerLinksFunc, NULL, reinterpret_cast<void*>(iPlayer));
}


void SimpleSend(int iSource, int iDest, const char* pszMessage, const cMultiParm& mpData1, const cMultiParm& mpData2, const cMultiParm& mpData3)
{
	cMultiParm mpReply;
	g_pScriptManager->SendMessage2(mpReply, iSource, iDest, pszMessage, mpData1, mpData2, mpData3);
	if (mpReply.type == kMT_String || mpReply.type == kMT_Vector)
	{
		g_pMalloc->Free(mpReply.psz);
	}
}

void SimplePost(int iSource, int iDest, const char* pszMessage, const cMultiParm& mpData1, const cMultiParm& mpData2, const cMultiParm& mpData3)
{
	g_pScriptManager->PostMessage2(iSource, iDest, pszMessage, mpData1, mpData2, mpData3);
}

long CDSend(const char* pszMessage, int iSource, const cMultiParm& mp)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	return pLS->BroadcastOnAllLinks(iSource, pszMessage, pLTS->LinkKindNamed("ControlDevice"), mp);
}


int ShowBook(int iObject, bool bArt, int iTime)
{
	SService<IPropertySrv> pPS(g_pScriptManager);
	if (!pPS->Possessed(iObject, "Book"))
		return 1;
	cMultiParm mpBook, mpBookArt;
	pPS->Get(mpBook, iObject, "Book", NULL);
	if (!*mpBook.psz)
	{
		g_pMalloc->Free(mpBook.psz);
		return 1;
	}
	SService<IDarkUISrv> pUI(g_pScriptManager);
	if (bArt && pPS->Possessed(iObject, "BookArt"))
	{
		pPS->Get(mpBookArt, iObject, "BookArt", NULL);
		pUI->ReadBook(mpBook, mpBookArt);
		g_pMalloc->Free(mpBookArt.psz);
		g_pMalloc->Free(mpBook.psz);
	}
	else
	{
		SService<IDataSrv> pDS(g_pScriptManager);
		char* pszBookFile = new char[strlen(mpBook) + 10];
		strcpy(pszBookFile, "..\\books\\");
		strcat(pszBookFile, mpBook);
		cScrStr str;
		pDS->GetString(str, pszBookFile, "page_0", "", "strings");
		if (str != NULL)
		{
			pUI->TextMessage(str, RGB(255,255,255), iTime);
			g_pMalloc->Free(const_cast<char*>(str.m_pszData));
		}
		delete[] pszBookFile;
		g_pMalloc->Free(mpBook.psz);
	}
	return 0;
}

void ShowString(const char* pszStr, int iTime, COLORREF color)
{
	SService<IDarkUISrv> pUI(g_pScriptManager);
	pUI->TextMessage(pszStr, color, iTime);
	return;
}

void DebugPrintf(const char* pszFormat, ...)
{
	va_list va;
	char *psz;
	int iLen;
	va_start(va, pszFormat);
	iLen = vsprintf(NULL, pszFormat, va);
	va_end(va);
	psz = new char[iLen+1];
	va_start(va, pszFormat);
	vsprintf(psz, pszFormat, va);
	va_end(va);

	IDebugScrSrv* pDSS = static_cast<IDebugScrSrv*>(g_pScriptManager->GetService(IID_IDebugScriptService));
	if (!pDSS)
	{
		delete[] psz;
		return;
	}
	pDSS->MPrint(cScrStr(psz), cScrStr(""), cScrStr(""), cScrStr(""), cScrStr(""), cScrStr(""), cScrStr(""), cScrStr(""));
	pDSS->Release();
#if DEBUG
	ShowString(psz, 1000 + (iLen * 50));
#endif
	delete[] psz;
}

void DisplayPrintf(const char* pszFormat, ...)
{
	va_list va;
	char *psz;
	int iLen;
	va_start(va, pszFormat);
	iLen = vsprintf(NULL, pszFormat, va);
	va_end(va);
	psz = new char[iLen+1];
	va_start(va, pszFormat);
	vsprintf(psz, pszFormat, va);
	va_end(va);

	ShowString(psz, 1000 + (iLen * 50));
	delete[] psz;
}


void MultiParmAssignInt(sMultiParm* pParm, int iValue)
{
	if (pParm->type == kMT_String || pParm->type == kMT_Vector)
	{
		g_pMalloc->Free(pParm->psz);
	}
	pParm->type = kMT_Int;
	pParm->i = iValue;
}

void MultiParmAssignFloat(sMultiParm* pParm, float fValue)
{
	if (pParm->type == kMT_String || pParm->type == kMT_Vector)
	{
		g_pMalloc->Free(pParm->psz);
	}
	pParm->type = kMT_Float;
	pParm->f = fValue;
}

void MultiParmAssignString(sMultiParm* pParm, const char* pszValue)
{
	if (pParm->type == kMT_String || pParm->type == kMT_Vector)
	{
		g_pMalloc->Free(pParm->psz);
		pParm->type = kMT_Undef;
	}
	pParm->psz = reinterpret_cast<char*>(g_pMalloc->Alloc(strlen(pszValue) + 1));
	if (pParm->psz)
	{
		strcpy(pParm->psz, pszValue);
		pParm->type = kMT_String;
	}
}

void MultiParmAssignVector(sMultiParm* pParm, const mxs_vector* pvValue)
{
	if (pParm->type == kMT_String || pParm->type == kMT_Vector)
	{
		g_pMalloc->Free(pParm->psz);
		pParm->type = kMT_Undef;
	}
	pParm->pVector = reinterpret_cast<mxs_vector*>(g_pMalloc->Alloc(sizeof(mxs_vector)));
	if (pParm->pVector)
	{
		memcpy(pParm->pVector, pvValue, sizeof(mxs_vector));
		pParm->type = kMT_Vector;
	}
}


