
#include <cstdarg>
#include <cstring>
#include <cstdio>
#include <cctype>
#include <vector>

#include "lg/interface.h"
#include "lg/scrservices.h"
#include "lg/scrmanagers.h"
#include "lg/objects.h"
#include "lg/links.h"

#include "ScriptLib.h"

#ifdef __BORLANDC__
#define strnicmp	std::strnicmp
#endif


extern int (__cdecl *g_pfnMPrintf)(const char*, ...);

IMalloc* g_pMalloc;
IScriptMan* g_pScriptManager;

const cMultiParm g_mpUndef;


/* Whoops. msvcrt doesn't include strsep.
 */
char* strsep(char** str, const char* delim)
{
	register char* p = *str;
	if (!p) return NULL;
	while (*p)
	{
		register const char* d = delim;
		while (*d)
		{
			if (*p == *d++)
			{
				register char* t = *str;
				*p++ = '\0';
				*str = p;
				return t;
			}
		}
		++p;
	}
	p = *str;
	*str = NULL;
	return p;
}

/* Seperate a string on any of delim characters while preserving 
 * substrings surrounded by an equal pair of any of quotes characters.
 * str is the address of a pointer to the string to be seperated.
 * The pointer to the next token in the string, or NULL, is copied to 
 * the address and the the value that was passed in *str is returned.
 * delim and quotes cannot be NULL.
 * Empty tokens will be returned.
 * Like strtok and strsep, the string data will be modified.
 * str, of course, must be an lvalue.
 */
char* strqsep(char** str, const char* delim, const char* quotes)
{
	register char* p = *str;
	if (!p) return NULL;
	while (*p)
	{
		{
			register const char* q = quotes;
			while (*q)
			{
				if (*p == *q++)
				{
					register char* t = p;
				L_ESCAPEQUOTE:
					t = strchr(t+1, *p);
					if (!t)
					{
						p = *str;
						*str = NULL;
						return p;
					}
					if (*(t-1) == '\\')
						goto L_ESCAPEQUOTE;
					p = t;
					goto L_QUOTE;
				}
			}
		}
		{
			register const char* d = delim;
			while (*d)
			{
				if (*p == *d++)
				{
					register char* t = *str;
					*p++ = '\0';
					*str = p;
					return t;
				}
			}
		}
	L_QUOTE:
		++p;
	}
	p = *str;
	*str = NULL;
	return p;
}

static const char g_cHexTable[] = {
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,		0,1,2,3,4,5,6,7,8,9,0,0,0,0,0,0,
	0,10,11,12,13,14,15,0,0,0,0,0,0,0,0,0,	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,10,11,12,13,14,15,0,0,0,0,0,0,0,0,0,	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
};

int strtocolor(const char* str)
{
	BYTE r,g,b;
	register unsigned char* c = reinterpret_cast<unsigned char*>(const_cast<char*>(str));
	if (*c != '#')
		return 0;
	if (strlen(reinterpret_cast<char*>(++c)) < 6)
		return 0;
	r  = g_cHexTable[*c++] << 4;
	r |= g_cHexTable[*c++];
	g  = g_cHexTable[*c++] << 4;
	g |= g_cHexTable[*c++];
	b  = g_cHexTable[*c++] << 4;
	b |= g_cHexTable[*c];
	return RGB(r,g,b);
}

int strtotime(const char* str)
{
	char* end = NULL;
	int t = strtoul(str, &end, 10);
	if (t > 0 && end)
	{
		if ((*end | 0x20) == 'm')
			t *= 60000;
		else if ((*end | 0x20) == 's')
			t *= 1000;
	}
	return t;
}


#if (_DARKGAME == 3)
#define OBJPARAM_PROPNAME	"ObjList"
#else
#define OBJPARAM_PROPNAME	"DesignNote"
#endif

char* GetObjectParams(int iObject)
{
	SService<IPropertySrv> pPS(g_pScriptManager);
	cMultiParm mpDesignNote;
	if (!pPS->Possessed(iObject, OBJPARAM_PROPNAME))
		return NULL;
	pPS->Get(mpDesignNote, iObject, OBJPARAM_PROPNAME, NULL);
	return (mpDesignNote.psz);
}

int SetObjectParams(int iObject, const char* pszParams)
{
	SService<IPropertySrv> pPS(g_pScriptManager);
	cMultiParm mpDesignNote = const_cast<char*>(pszParams);
	pPS->Add(iObject, OBJPARAM_PROPNAME);
	return pPS->Set(iObject, OBJPARAM_PROPNAME, mpDesignNote);
}

char* GetObjectParamString(int iObject, const char* pszParam)
{
	char* pszString = GetObjectParams(iObject);
	char* pszRet = GetParamString(pszString, pszParam);
	g_pMalloc->Free(pszString);
	return pszRet;
}

int GetObjectParamInt(int iObject, const char* pszParam)
{
	char* params = GetObjectParams(iObject);
	if (!params)
		return 0;
	int iRet = GetParamInt(params, pszParam);
	g_pMalloc->Free(params);
	return iRet;
}

float GetObjectParamFloat(int iObject, const char* pszParam)
{
	char* params = GetObjectParams(iObject);
	if (!params)
		return 0;
	float fRet = GetParamFloat(params, pszParam);
	g_pMalloc->Free(params);
	return fRet;
}

int SetObjectParamString(int iObject, const char* pszParam, const char* pszVal)
{
	if (!pszParam || !pszVal)
		return 1;
	char* params = GetObjectParams(iObject);
	if (!params)
	{
		int iLen;
		// You'll have to take care not to pass a string with un-escaped quotes.
		iLen = sprintf(NULL, "%s=\"%s\"", pszParam, pszVal);
		params = reinterpret_cast<char*>(g_pMalloc->Alloc(iLen + 1));
		if (!params)
			return 1;
		sprintf(params, "%s=\"%s\"", pszParam, pszVal);
	}
	else
	{
		char* newparams = SetParamString(params, pszParam, pszVal);
		g_pMalloc->Free(params);
		params = newparams;
	}
	int iRet = 1;
	if (params)
	{
		iRet = SetObjectParams(iObject, params);
		g_pMalloc->Free(params);
	}
	return iRet;
}

int SetObjectParamInt(int iObject, const char* pszParam, int iVal)
{
	if (!pszParam)
		return 1;
	char* params = GetObjectParams(iObject);
	if (!params)
	{
		int iLen = sprintf(NULL, "%s=%d", pszParam, iVal);
		params = reinterpret_cast<char*>(g_pMalloc->Alloc(iLen + 1));
		if (!params)
			return 1;
		sprintf(params, "%s=%d", pszParam, iVal);
	}
	else
	{
		char* newparams = SetParamInt(params, pszParam, iVal);
		g_pMalloc->Free(params);
		params = newparams;
	}
	int iRet = 1;
	if (params)
	{
		iRet = SetObjectParams(iObject, params);
		g_pMalloc->Free(params);
	}
	return iRet;
}

int SetObjectParamFloat(int iObject, const char* pszParam, float fVal)
{
	if (!pszParam)
		return 1;
	char* params = GetObjectParams(iObject);
	if (!params)
	{
		int iLen = sprintf(NULL, "%s=%0.2f", pszParam, fVal);
		params = reinterpret_cast<char*>(g_pMalloc->Alloc(iLen + 1));
		if (!params)
			return 1;
		sprintf(params, "%s=%0.2f", pszParam, fVal);
	}
	else
	{
		char* newparams = SetParamFloat(params, pszParam, fVal);
		g_pMalloc->Free(params);
		params = newparams;
	}
	int iRet = 1;
	if (params)
	{
		iRet = SetObjectParams(iObject, params);
		g_pMalloc->Free(params);
	}
	return iRet;
}

char* GetParamString(const char* pszString, const char* pszParam)
{
	if (!pszString || !pszParam)
		return NULL;
	char* pszRet = NULL;
	char* pszTemp = new char[strlen(pszString) + 1];
	strcpy(pszTemp, pszString);
	char* pszSep, *pszToken;
	for (pszSep = pszTemp, pszToken = strqsep(&pszSep, ";", "\"'"); 
	     pszToken; pszToken = strqsep(&pszSep, ";", "\"'"))
	{
		while (isspace(*pszToken)) ++pszToken;
		char* pszStart = strchr(pszToken, '=');
		if (pszStart)
			*pszStart++ = '\0';
		if (!stricmp(pszToken, pszParam))
		{
			if (!pszStart)
				break;
			while (isspace(*pszStart)) ++pszStart;
			char* pszEnd;
			if (*pszStart == '\"')
			{
				++pszStart;
				pszEnd = strrchr(pszStart, '\"');
				if (!pszEnd)
					pszEnd = pszStart + strlen(pszStart);
			}
			else if (*pszStart == '\'')
			{
				++pszStart;
				pszEnd = strrchr(pszStart, '\'');
				if (!pszEnd)
					pszEnd = pszStart + strlen(pszStart);
			}
			else
				pszEnd = pszStart + strlen(pszStart);
			int iLen = pszEnd - pszStart;
			pszRet = reinterpret_cast<char*>(g_pMalloc->Alloc(iLen + 1));
			if (pszRet)
			{
				strncpy(pszRet, pszStart, iLen);
				pszRet[iLen] = '\0';
			}
			break;
		}
	}
	delete[] pszTemp;
	return pszRet;
}

int GetParamInt(const char* pszString, const char* pszParam)
{
	if (!pszString || !pszParam)
		return 0;
	int iRet = 0;
	char* pszTemp = new char[strlen(pszString) + 1];
	strcpy(pszTemp, pszString);
	char* pszSep, *pszToken;
	for (pszSep = pszTemp, pszToken = strqsep(&pszSep, ";", "\"'"); 
	     pszToken; pszToken = strqsep(&pszSep, ";", "\"'"))
	{
		while (isspace(*pszToken)) ++pszToken;
		char* pszStart = strchr(pszToken, '=');
		if (pszStart)
			*pszStart++ = '\0';
		if (!stricmp(pszToken, pszParam))
		{
			if (!pszStart)
				break;
			while (isspace(*pszStart)) ++pszStart;
			char* pszEnd = pszStart + strlen(pszStart);
			if (*pszStart == '\"' || *pszStart == '\'')
			{
				if (*(pszEnd-1) == *pszStart)
					*(--pszEnd) = '\0';
				++pszStart;
			}
			if (*pszStart == '$')
			{
				++pszStart;
				if (pszEnd > pszStart)
				{
					SService<IQuestSrv> pQS(g_pScriptManager);
					if (pQS->Exists(pszStart))
						iRet = pQS->Get(pszStart);
				}
			}
			else
			{
				iRet = strtol(pszStart, NULL, 0);
			}
			break;
		}
	}
	delete[] pszTemp;
	return iRet;
}

float GetParamFloat(const char* pszString, const char* pszParam)
{
	if (!pszString || !pszParam)
		return 0;
	float fRet = 0;
	char* pszTemp = new char[strlen(pszString) + 1];
	strcpy(pszTemp, pszString);
	char* pszSep, *pszToken;
	for (pszSep = pszTemp, pszToken = strqsep(&pszSep, ";", "\"'"); 
	     pszToken; pszToken = strqsep(&pszSep, ";", "\"'"))
	{
		while (isspace(*pszToken)) ++pszToken;
		char* pszStart = strchr(pszToken, '=');
		if (pszStart)
			*pszStart++ = '\0';
		if (!stricmp(pszToken, pszParam))
		{
			if (!pszStart)
				break;
			while (isspace(*pszStart)) ++pszStart;
			if (*pszStart == '\"' || *pszStart == '\'')
				++pszStart;
			fRet = strtod(pszStart, NULL);
			break;
		}
	}
	delete[] pszTemp;
	return fRet;
}

char* SetParamString(char* pszString, const char* pszParam, const char* pszVal)
{
	if (!pszString || !pszParam || !pszVal)
		return NULL;
	int iLen = strlen(pszString) + strlen(pszParam) + strlen(pszVal) + 4;
	char* pszNew = reinterpret_cast<char*>(g_pMalloc->Alloc(iLen + 1));
	if (!pszNew)
		return NULL;
	sprintf(pszNew, "%s=\"%s\";", pszParam, pszVal);
	char* pszSep, *pszToken;
	for (pszSep = pszString, pszToken = strqsep(&pszSep, ";", "\"'"); 
	     pszToken; pszToken = strqsep(&pszSep, ";", "\"'"))
	{
		while (isspace(*pszToken)) ++pszToken;
		if (!*pszToken)
			continue;
		char* pszStart = strchr(pszToken, '=');
		if (pszStart)
			*pszStart++ = '\0';
		if (stricmp(pszToken, pszParam))
		{
			strcat(pszNew, pszToken);
			if (pszStart)
			{
				strcat(pszNew, "=");
				strcat(pszNew, pszStart);
			}
			strcat(pszNew, ";");
		}
	}
	return pszNew;
}

char* SetParamInt(char* pszString, const char* pszParam, int iVal)
{
	if (!pszString || !pszParam)
		return NULL;
	int iLen = strlen(pszString) + sprintf(NULL, "%s=%d;", pszParam, iVal);
	char* pszNew = reinterpret_cast<char*>(g_pMalloc->Alloc(iLen + 1));
	if (!pszNew)
		return NULL;
	sprintf(pszNew, "%s=%d;", pszParam, iVal);
	char* pszSep, *pszToken;
	for (pszSep = pszString, pszToken = strqsep(&pszSep, ";", "\"'"); 
	     pszToken; pszToken = strqsep(&pszSep, ";", "\"'"))
	{
		while (isspace(*pszToken)) ++pszToken;
		if (!*pszToken)
			continue;
		char* pszStart = strchr(pszToken, '=');
		if (pszStart)
			*pszStart++ = '\0';
		if (stricmp(pszToken, pszParam))
		{
			strcat(pszNew, pszToken);
			if (pszStart)
			{
				strcat(pszNew, "=");
				strcat(pszNew, pszStart);
			}
			strcat(pszNew, ";");
		}
	}
	return pszNew;
}

char* SetParamFloat(char* pszString, const char* pszParam, float fVal)
{
	if (!pszString || !pszParam)
		return NULL;
	int iLen = strlen(pszString) + sprintf(NULL, "%s=%0.2f;", pszParam, fVal);
	char* pszNew = reinterpret_cast<char*>(g_pMalloc->Alloc(iLen + 1));
	if (!pszNew)
		return NULL;
	sprintf(pszNew, "%s=%0.2f;", pszParam, fVal);
	char* pszSep, *pszToken;
	for (pszSep = pszString, pszToken = strqsep(&pszSep, ";", "\"'"); 
	     pszToken; pszToken = strqsep(&pszSep, ";", "\"'"))
	{
		while (isspace(*pszToken)) ++pszToken;
		if (!*pszToken)
			continue;
		char* pszStart = strchr(pszToken, '=');
		if (pszStart)
			*pszStart++ = '\0';
		if (stricmp(pszToken, pszParam))
		{
			strcat(pszNew, pszToken);
			if (pszStart)
			{
				strcat(pszNew, "=");
				strcat(pszNew, pszStart);
			}
			strcat(pszNew, ";");
		}
	}
	return pszNew;
}

int RemoveObjectParam(int iObject, const char* pszParam)
{
	char* pszString = GetObjectParams(iObject);
	if (!pszString)
		return 0;
	char* pszNew = new char[strlen(pszString) + 1];
	*pszNew = '\0';
	char* pszSep, *pszToken;
	for (pszSep = pszString, pszToken = strqsep(&pszSep, ";", "\"'"); 
	     pszToken; pszToken = strqsep(&pszSep, ";", "\"'"))
	{
		while (isspace(*pszToken)) ++pszToken;
		if (!*pszToken)
			continue;
		char* pszStart = strchr(pszToken, '=');
		if (pszStart)
			*pszStart++ = '\0';
		if (stricmp(pszToken, pszParam))
		{
			strcat(pszNew, pszToken);
			if (pszStart)
			{
				strcat(pszNew, "=");
				strcat(pszNew, pszStart);
			}
			strcat(pszNew, ";");
		}
	}
	SetObjectParams(iObject, pszNew);
	delete[] pszNew;
	g_pMalloc->Free(pszString);
	return 0;
}

int GetObjectParamObject(int iObject, const char* pszParam)
{
	char* params = GetObjectParams(iObject);
	if (!params)
		return 0;
	int iRet = GetParamObject(params, pszParam);
	g_pMalloc->Free(params);
	return iRet;
}

int GetParamObject(const char* pszString, const char* pszParam)
{
	if (!pszString || !pszParam)
		return 0;
	char* pszRet = NULL;
	char* pszTemp = new char[strlen(pszString) + 1];
	strcpy(pszTemp, pszString);
	char* pszSep, *pszToken;
	for (pszSep = pszTemp, pszToken = strqsep(&pszSep, ";", "\"'"); 
	     pszToken; pszToken = strqsep(&pszSep, ";", "\"'"))
	{
		while (isspace(*pszToken)) ++pszToken;
		char* pszStart = strchr(pszToken, '=');
		if (pszStart)
			*pszStart++ = '\0';
		if (!stricmp(pszToken, pszParam))
		{
			if (!pszStart)
				break;
			while (isspace(*pszStart)) ++pszStart;
			char* pszEnd;
			if (*pszStart == '\"')
			{
				++pszStart;
				pszEnd = strrchr(pszStart, '\"');
				if (!pszEnd)
					pszEnd = pszStart + strlen(pszStart);
			}
			else if (*pszStart == '\'')
			{
				++pszStart;
				pszEnd = strrchr(pszStart, '\'');
				if (!pszEnd)
					pszEnd = pszStart + strlen(pszStart);
			}
			else
				pszEnd = pszStart + strlen(pszStart);
			int iLen = pszEnd - pszStart;
			pszRet = pszStart;
			pszRet[iLen] = '\0';
			break;
		}
	}
	int iRet = StrToObject(pszRet);
	delete[] pszTemp;
	return iRet;
}

int GetObjectParamTime(int iObject, const char* pszParam)
{
	char* params = GetObjectParams(iObject);
	if (!params)
		return 0;
	int iRet = GetParamTime(params, pszParam);
	g_pMalloc->Free(params);
	return iRet;
}

int GetParamTime(const char* pszString, const char* pszParam)
{
	if (!pszString || !pszParam)
		return 0;
	int iRet = 0;
	char* pszTemp = new char[strlen(pszString) + 1];
	strcpy(pszTemp, pszString);
	char* pszSep, *pszToken;
	for (pszSep = pszTemp, pszToken = strqsep(&pszSep, ";", "\"'"); 
	     pszToken; pszToken = strqsep(&pszSep, ";", "\"'"))
	{
		while (isspace(*pszToken)) ++pszToken;
		char* pszStart = strchr(pszToken, '=');
		if (pszStart)
			*pszStart++ = '\0';
		if (!stricmp(pszToken, pszParam))
		{
			if (!pszStart)
				break;
			while (isspace(*pszStart)) ++pszStart;
			char* pszEnd = pszStart + strlen(pszStart);
			if (*pszStart == '\"' || *pszStart == '\'')
			{
				if (*(pszEnd-1) == *pszStart)
					*(--pszEnd) = '\0';
				++pszStart;
			}
			if (*pszStart == '$')
			{
				++pszStart;
				if (pszEnd > pszStart)
				{
					SService<IQuestSrv> pQS(g_pScriptManager);
					if (pQS->Exists(pszStart))
						iRet = pQS->Get(pszStart);
				}
			}
			else
			{
				iRet = strtotime(pszStart);
			}
			break;
		}
	}
	delete[] pszTemp;
	return iRet;
}


static int DoAddMetaProperty(IObjectSrv* pOS, object oMetaProp, object oObject)
{
	int iRet = 0;
	true_bool has_mp;
	pOS->HasMetaProperty(has_mp, oObject, oMetaProp);
	if (!has_mp)
	{
		iRet = pOS->AddMetaProperty(oObject, oMetaProp);
	}
	return iRet;
}

static int DoRemoveMetaProperty(IObjectSrv* pOS, object oMetaProp, object oObject)
{
	int iRet = 0;
	true_bool has_mp;
	pOS->HasMetaProperty(has_mp, oObject, oMetaProp);
	if (has_mp)
	{
		iRet = pOS->RemoveMetaProperty(oObject, oMetaProp);
		//pOS->HasMetaProperty(has_mp, oObject, oMetaProp);
	}
	return iRet;
}

int AddSingleMetaProperty(const char* pszMetaProp, int iObjId)
{
	SService<IObjectSrv> pOS(g_pScriptManager);
	object oMetaProp;
	pOS->Named(oMetaProp, pszMetaProp);
	if (oMetaProp != 0)
		return DoAddMetaProperty(pOS, oMetaProp, iObjId);
	return 0;
}

int AddSingleMetaProperty(int iMetaProp, int iObjId)
{
	SService<IObjectSrv> pOS(g_pScriptManager);
	return DoAddMetaProperty(pOS, iMetaProp, iObjId);
}

int RemoveSingleMetaProperty(const char* pszMetaProp, int iObjId)
{
	SService<IObjectSrv> pOS(g_pScriptManager);
	object oMetaProp;
	pOS->Named(oMetaProp, pszMetaProp);
	if (oMetaProp != 0)
		return DoRemoveMetaProperty(pOS, oMetaProp, iObjId);
	return 0;
}

int RemoveSingleMetaProperty(int iMetaProp, int iObjId)
{
	SService<IObjectSrv> pOS(g_pScriptManager);
	return DoRemoveMetaProperty(pOS, iMetaProp, iObjId);
}


long GetOneLinkInheritedSrc(const char* pszFlavor, int iSource, int iDest, sLink* pResult)
{
	SInterface<IObjectSystem> pOS(g_pScriptManager);
	SInterface<ITraitManager> pTM(g_pScriptManager);
	SInterface<ILinkManager> pLM(g_pScriptManager);
	SInterface<IRelation> pRel;
	pRel = pLM->GetRelationNamed(pszFlavor);
	if (!pRel)
		return 0;

	long iLinkID = 0;
	int iObj = iSource;
	while (pOS->Exists(iObj))
	{
		if (pRel->AnyLinks(iObj, iDest))
		{
			ILinkQuery* pLQ = pRel->Query(iObj, iDest);
			iLinkID = pLQ->ID();
			pLQ->Release();
			break;
		}
		iObj = pTM->GetArchetype(iObj);
	}
	if (iLinkID && pResult)
		pLM->Get(iLinkID, pResult);
	return iLinkID;
}

int GetOneLinkDest(const char* pszFlavor, int iSource)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	link l;
	linkkind lkFlavor = (pszFlavor) ? pLTS->LinkKindNamed(pszFlavor) : 0;
	pLS->GetOne(l,lkFlavor, iSource, 0);
	if (l == 0)
		return 0;
	sLink sl;
	pLTS->LinkGet(l, sl);
	return sl.dest;
}

long GetAnyLink(const char* pszFlavor, int iSource, int iDest, sLink* pResult)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);

	std::vector<long> vLinks;
	linkset lsLinks;
	linkkind lkFlavor = (pszFlavor) ? pLTS->LinkKindNamed(pszFlavor) : 0;
	pLS->GetAll(lsLinks, lkFlavor, iSource, iDest);
	for (true_bool __p; lsLinks.AnyLinksLeft(__p), __p; lsLinks.NextLink())
	{
		vLinks.insert(vLinks.begin(), lsLinks.query->ID());
	}

	long iLinkID = 0;
	if (!vLinks.empty())
	{
		SService<IDataSrv> pDS(g_pScriptManager);
		iLinkID = vLinks.at(pDS->RandInt(0, vLinks.size() - 1));
		if (iLinkID && pResult)
			pLTS->LinkGet(iLinkID, *pResult);
	}
	return iLinkID;
}

long GetAnyLinkInheritedSrc(const char* pszFlavor, int iSource, int iDest, sLink* pResult)
{
	SInterface<IObjectSystem> pOS(g_pScriptManager);
	SInterface<ITraitManager> pTM(g_pScriptManager);
	SInterface<ILinkManager> pLM(g_pScriptManager);
	SInterface<IRelation> pRel;
	pRel = pLM->GetRelationNamed(pszFlavor);
	if (!pRel)
		return 0;

	std::vector<long> vLinks;
	int iObj = iSource;
	while (pOS->Exists(iObj))
	{
		if (pRel->AnyLinks(iObj, iDest))
		{
			ILinkQuery* pLQ = pRel->Query(iObj, iDest);
			for (; ! pLQ->Done(); pLQ->Next())
			{
				vLinks.insert(vLinks.begin(), pLQ->ID());
			}
			pLQ->Release();
		}
		iObj = pTM->GetArchetype(iObj);
	}

	long iLinkID = 0;
	if (!vLinks.empty())
	{
		SService<IDataSrv> pDS(g_pScriptManager);
		iLinkID = vLinks.at(pDS->RandInt(0, vLinks.size() - 1));
		if (iLinkID && pResult)
			pLM->Get(iLinkID, pResult);
	}
	return iLinkID;
}

int GetAnyLinkDest(const char* pszFlavor, int iSource)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);

	std::vector<long> vLinks;
	linkset lsLinks;
	linkkind lkFlavor = (pszFlavor) ? pLTS->LinkKindNamed(pszFlavor) : 0;
	pLS->GetAll(lsLinks, lkFlavor, iSource, 0);
	for (true_bool __p; lsLinks.AnyLinksLeft(__p), __p; lsLinks.NextLink())
	{
		vLinks.insert(vLinks.begin(), lsLinks.query->ID());
	}

	if (vLinks.empty())
		return 0;

	SService<IDataSrv> pDS(g_pScriptManager);
	sLink sl;
	pLTS->LinkGet(vLinks.at(pDS->RandInt(0, vLinks.size() - 1)), sl);
	return sl.dest;
}

long GetOneLinkByData(const char* pszFlavor, int iSource, int iDest, sLink* pResult, const void* pvData, int iDataLen)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);

	int __cdecl (*cmp)(const void*, const void*, size_t) = memcmp;
	if (pvData && iDataLen == -1)
	{
		iDataLen = strlen(reinterpret_cast<const char*>(pvData)) + 1;
		cmp = reinterpret_cast<int (*)(const void*,const void*,size_t)>(strnicmp);
	}

	linkset lsLinks;
	linkkind lkFlavor = (pszFlavor) ? pLTS->LinkKindNamed(pszFlavor) : 0;
	pLS->GetAll(lsLinks, lkFlavor, iSource, iDest);
	long iLinkID = 0;
	for (true_bool __p; lsLinks.AnyLinksLeft(__p), __p; lsLinks.NextLink())
	{
		void* pLinkData = lsLinks.query->Data();
		if (pvData)
		{
			if (!pLinkData)
				continue;
			if ((*cmp)(pvData, pLinkData, iDataLen))
				continue;
		}
		else
		{
			if (pLinkData && *reinterpret_cast<const char*>(pLinkData))
				continue;
		}
		iLinkID = lsLinks.query->ID();
		if (pResult)
			lsLinks.query->Link(pResult);
	}
	return iLinkID;
}

long GetOneLinkByDataInheritedSrc(const char* pszFlavor, int iSource, int iDest, sLink* pResult, const void* pvData, int iDataLen)
{
	SInterface<IObjectSystem> pOS(g_pScriptManager);
	SInterface<ITraitManager> pTM(g_pScriptManager);
	SInterface<ILinkManager> pLM(g_pScriptManager);
	SInterface<IRelation> pRel;
	pRel = pLM->GetRelationNamed(pszFlavor);
	if (!pRel)
		return 0;

	int __cdecl (*cmp)(const void*, const void*, size_t) = memcmp;
	if (pvData && iDataLen == -1)
	{
		iDataLen = strlen(reinterpret_cast<const char*>(pvData)) + 1;
		cmp = reinterpret_cast<int (*)(const void*,const void*,size_t)>(strnicmp);
	}

	long iLinkID = 0;
	int iObj = iSource;
	while (pOS->Exists(iObj))
	{
		if (pRel->AnyLinks(iObj, iDest))
		{
			ILinkQuery* pLQ = pRel->Query(iObj, iDest);
			for (; ! pLQ->Done(); pLQ->Next())
			{
				void* pLinkData = pLQ->Data();
				if (pvData)
				{
					if (pLinkData && !(*cmp)(pvData, pLinkData, iDataLen))
					{
						iLinkID = pLQ->ID();
						if (pResult)
							pLQ->Link(pResult);
						break;
					}
				}
				else
				{
					if (!pLinkData || !*reinterpret_cast<const char*>(pLinkData))
						break;
				}
			}
			pLQ->Release();
		}
		if (iLinkID != 0)
			break;
		iObj = pTM->GetArchetype(iObj);
	}
	return iLinkID;
}

long GetAnyLinkByData(const char* pszFlavor, int iSource, int iDest, sLink* pResult, const void* pvData, int iDataLen)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);

	int __cdecl (*cmp)(const void*, const void*, size_t) = memcmp;
	if (pvData && iDataLen == -1)
	{
		iDataLen = strlen(reinterpret_cast<const char*>(pvData)) + 1;
		cmp = reinterpret_cast<int (*)(const void*,const void*,size_t)>(strnicmp);
	}

	std::vector<long> vLinks;
	linkset lsLinks;
	linkkind lkFlavor = (pszFlavor) ? pLTS->LinkKindNamed(pszFlavor) : 0;
	pLS->GetAll(lsLinks, lkFlavor, iSource, iDest);
	for (true_bool __p; lsLinks.AnyLinksLeft(__p), __p; lsLinks.NextLink())
	{
		void* pLinkData = lsLinks.query->Data();
		if (pvData)
		{
			if (!pLinkData)
				continue;
			if ((*cmp)(pvData, pLinkData, iDataLen))
				continue;
		}
		else
		{
			if (pLinkData && *reinterpret_cast<const char*>(pLinkData))
				continue;
		}
		vLinks.insert(vLinks.begin(), lsLinks.query->ID());
	}

	long iLinkID = 0;
	if (!vLinks.empty())
	{
		SService<IDataSrv> pDS(g_pScriptManager);
		iLinkID = vLinks.at(pDS->RandInt(0, vLinks.size() - 1));
		if (iLinkID && pResult)
			pLTS->LinkGet(iLinkID, *pResult);
	}
	return iLinkID;
}

long GetAnyLinkByDataInheritedSrc(const char* pszFlavor, int iSource, int iDest, sLink* pResult, const void* pvData, int iDataLen)
{
	SInterface<IObjectSystem> pOS(g_pScriptManager);
	SInterface<ITraitManager> pTM(g_pScriptManager);
	SInterface<ILinkManager> pLM(g_pScriptManager);
	SInterface<IRelation> pRel;
	pRel = pLM->GetRelationNamed(pszFlavor);
	if (!pRel)
		return 0;

	int __cdecl (*cmp)(const void*, const void*, size_t) = memcmp;
	if (pvData && iDataLen == -1)
	{
		iDataLen = strlen(reinterpret_cast<const char*>(pvData)) + 1;
		cmp = reinterpret_cast<int (*)(const void*,const void*,size_t)>(strnicmp);
	}

	std::vector<long> vLinks;
	int iObj = iSource;
	while (pOS->Exists(iObj))
	{
		if (pRel->AnyLinks(iObj, iDest))
		{
			ILinkQuery* pLQ = pRel->Query(iObj, iDest);
			for (; ! pLQ->Done(); pLQ->Next())
			{
				void* pLinkData = pLQ->Data();
				if (pvData)
				{
					if (pLinkData && !(*cmp)(pvData, pLinkData, iDataLen))
					{
						vLinks.insert(vLinks.begin(), pLQ->ID());
					}
				}
				else
				{
					if (!pLinkData || !*reinterpret_cast<const char*>(pLinkData))
					{
						vLinks.insert(vLinks.begin(), pLQ->ID());
					}
				}
			}
			pLQ->Release();
		}
		if (!vLinks.empty())
			break;
		iObj = pTM->GetArchetype(iObj);
	}

	long iLinkID = 0;
	if (!vLinks.empty())
	{
		SService<IDataSrv> pDS(g_pScriptManager);
		iLinkID = vLinks.at(pDS->RandInt(0, vLinks.size() - 1));
		if (iLinkID && pResult)
			pLM->Get(iLinkID, pResult);
	}
	return iLinkID;
}

long IterateLinks(const char* pszFlavor, int iSource, int iDest,
			    LinkIterCallbackFunc pfnLinkIterCallback, IScript* pScript, void* pData)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);

	linkset lsLinks;
	linkkind lkFlavor = (pszFlavor) ? pLTS->LinkKindNamed(pszFlavor) : 0;
	pLS->GetAll(lsLinks, lkFlavor, iSource, iDest);
	long iCount = 0;
	for (true_bool __p; lsLinks.AnyLinksLeft(__p), __p; lsLinks.NextLink())
	{
		++iCount;
		// It might be worthwhile to wrap this in a try .. catch
		if (!pfnLinkIterCallback(pLS, lsLinks.query, pScript, pData))
			break;
	}
	return iCount;
}

long IterateLinksByData(const char* pszFlavor, int iSource, int iDest,
			    const void* pvFilter, int iFilterLen,
			    LinkIterCallbackFunc pfnLinkIterCallback, IScript* pScript, void* pData)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);

	int __cdecl (*cmp)(const void*, const void*, size_t) = memcmp;
	if (pvFilter && iFilterLen == -1)
	{
		iFilterLen = strlen(reinterpret_cast<const char*>(pvFilter)) + 1;
		cmp = reinterpret_cast<int (*)(const void*,const void*,size_t)>(strnicmp);
	}

	linkset lsLinks;
	linkkind lkFlavor = (pszFlavor) ? pLTS->LinkKindNamed(pszFlavor) : 0;
	pLS->GetAll(lsLinks, lkFlavor, iSource, iDest);
	long iCount = 0;
	for (true_bool __p; lsLinks.AnyLinksLeft(__p), __p; lsLinks.NextLink())
	{
		void* pLinkData = lsLinks.query->Data();
		if (pvFilter)
		{
			if (!pLinkData)
				continue;
			if ((*cmp)(pvFilter, pLinkData, iFilterLen))
				continue;
		}
		else
		{
			if (pLinkData && *reinterpret_cast<const char*>(pLinkData))
				continue;
		}
		++iCount;
		// It might be worthwhile to wrap this in a try .. catch
		if (!pfnLinkIterCallback(pLS, lsLinks.query, pScript, pData))
			break;
	}
	return iCount;
}

void DestroyAllLinksByData(const char* pszFlavor, int iSource, int iDest,
			    const void* pvFilter, int iFilterLen)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);

	int __cdecl (*cmp)(const void*, const void*, size_t) = memcmp;
	if (pvFilter && iFilterLen == -1)
	{
		iFilterLen = strlen(reinterpret_cast<const char*>(pvFilter)) + 1;
		cmp = reinterpret_cast<int (*)(const void*,const void*,size_t)>(strnicmp);
	}

	linkset lsLinks;
	linkkind lkFlavor = (pszFlavor) ? pLTS->LinkKindNamed(pszFlavor) : 0;
	pLS->GetAll(lsLinks, lkFlavor, iSource, iDest);
	for (true_bool __p; lsLinks.AnyLinksLeft(__p), __p; lsLinks.NextLink())
	{
		void* pLinkData = lsLinks.query->Data();
		if (pvFilter)
		{
			if (!pLinkData)
				continue;
			if ((*cmp)(pvFilter, pLinkData, iFilterLen))
				continue;
		}
		else
		{
			if (pLinkData && *reinterpret_cast<const char*>(pLinkData))
				continue;
		}
		pLS->Destroy(lsLinks.query->ID());
	}
}

static int FixupPlayerLinksFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	// Move the destination of the link to the object specified in pData
	SInterface<ILinkManager> pLM(g_pScriptManager);
	int iPlayer = reinterpret_cast<int>(pData);
	sLink slInfo;
	pLQ->Link(&slInfo);
	pLM->AddFull(slInfo.source, iPlayer, slInfo.flavor, pLQ->Data());
	pLM->Remove(pLQ->ID());
	return 1;
}

void FixupPlayerLinks(int iSource, int iPlayer)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	// Get the PlayerFactory link. There should only ever be one of these.
	link lPFLink;
	pLS->GetOne(lPFLink, pLTS->LinkKindNamed("PlayerFactory"), 0, 0);
	if (lPFLink == 0)
		return;
	// Iterate through all links from here to the source of the PF link.
	sLink lPFInfo;
	pLTS->LinkGet(lPFLink, lPFInfo);
	IterateLinks(NULL, iSource, lPFInfo.source, FixupPlayerLinksFunc, NULL, reinterpret_cast<void*>(iPlayer));
}


void SimpleSend(int iSource, int iDest, const char* pszMessage, const cMultiParm& mpData1, const cMultiParm& mpData2, const cMultiParm& mpData3)
{
	cMultiParm mpReply;
	g_pScriptManager->SendMessage2(mpReply, iSource, iDest, pszMessage, mpData1, mpData2, mpData3);
	mpReply.Free();
}

void SimplePost(int iSource, int iDest, const char* pszMessage, const cMultiParm& mpData1, const cMultiParm& mpData2, const cMultiParm& mpData3)
{
	g_pScriptManager->PostMessage2(iSource, iDest, pszMessage, mpData1, mpData2, mpData3);
}

#if (_DARKGAME == 1)
long CDSend(const char* pszMessage, int iSource)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	return pLS->BroadcastOnAllLinks(iSource, pszMessage, pLTS->LinkKindNamed("ControlDevice"));
}
long CDSend(const char* pszMessage, int iSource, const cMultiParm& mp)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	linkset lsLinks;
	pLS->GetAll(lsLinks, pLTS->LinkKindNamed("ControlDevice"), iSource, 0);
	for (true_bool __p; lsLinks.AnyLinksLeft(__p), __p; lsLinks.NextLink())
	{
		sLink sl;
		lsLinks.query->Link(&sl);
		g_pScriptManager->PostMessage2(iSource, sl.dest, pszMessage, mp, g_mpUndef, g_mpUndef);
	}
	return 0;
}
#else
long CDSend(const char* pszMessage, int iSource, const cMultiParm& mp)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	return pLS->BroadcastOnAllLinks(iSource, pszMessage, 
#if (_DARKGAME == 3)
		pLTS->LinkKindNamed("SwitchLink"),
#else
		pLTS->LinkKindNamed("ControlDevice"),
#endif
		mp);
}
#endif


int StrToObject(const char* pszName)
{
	if (!pszName || !*pszName)
		return 0;
	SService<IObjectSrv> pOS(g_pScriptManager);
	object obj = 0;
	pOS->Named(obj, pszName);
	if (obj == 0)
	{
		char *p = NULL;
		obj = strtol(pszName, &p, 10);
		//if (*p != '\0')
		//	return 0;
		true_bool _p;
		if (! *pOS->Exists(_p, obj))
			return 0;
	}
	return obj;
}


int CalcTextTime(const char* pszText, int iWordTime)
{
	register const char *p = pszText;
	register const char *q = p;
	register int c = 0, e = 0;
	while (*p)
	{
		if (isspace(*p))
		{
			if ((p - q) + e > 3)
			{
				++c;
				e = 0;
			}
			else
				++e;
			while (isspace(*++p)) ;
			q = p;
			continue;
		}
		++p;
	}
	if ((p - q) + e > 3)
		++c;
	//if (c < ((p-pszText)/8))
	//	c = (p-pszText)/8;
	return iWordTime * ((c < 10) ? 10 : c);
}

int ShowBook(int iObject, bool bArt, int iTime)
{
#if (_DARKGAME == 3)
	SService<IShockGameSrv> pShock(g_pScriptManager);
	pShock->UseLog(iObject, 1);
#else
	SService<IPropertySrv> pPS(g_pScriptManager);
	if (!pPS->Possessed(iObject, "Book"))
		return 1;
	cMultiParm mpBook, mpBookArt;
	pPS->Get(mpBook, iObject, "Book", NULL);
	if (!*mpBook.psz)
	{
		g_pMalloc->Free(mpBook.psz);
		return 1;
	}
	SService<IDarkUISrv> pUI(g_pScriptManager);
	if (bArt && pPS->Possessed(iObject, "BookArt"))
	{
		pPS->Get(mpBookArt, iObject, "BookArt", NULL);
		pUI->ReadBook(mpBook, mpBookArt);
		g_pMalloc->Free(mpBookArt.psz);
		g_pMalloc->Free(mpBook.psz);
	}
	else
	{
		SService<IDataSrv> pDS(g_pScriptManager);
		char* pszBookFile = new char[strlen(mpBook) + 10];
		strcpy(pszBookFile, "..\\books\\");
		strcat(pszBookFile, mpBook);
		cScrStr str;
		pDS->GetString(str, pszBookFile, "page_0", "", "strings");
		if (*static_cast<const char*>(str))
		{
			if (iTime == 0)
				iTime = CalcTextTime(str, 500);
			pUI->TextMessage(str, 0, iTime);
		}
		str.Free();
		delete[] pszBookFile;
		g_pMalloc->Free(mpBook.psz);
	}
#endif
	return 0;
}

void ShowString(const char* pszStr, int iTime, COLORREF color)
{
#if (_DARKGAME == 3)
	SService<IShockGameSrv> pShock(g_pScriptManager);
	pShock->AddText(pszStr, 0, iTime);
#else
	SService<IDarkUISrv> pUI(g_pScriptManager);
	pUI->TextMessage(pszStr, color, iTime);
#endif
	return;
}

// vs[n]printf is pre-C99 in msvcrt6, 
// so we can't dynamically allocating a buffer.
// 1000 is a limit imposed by Dromed's mprint 
void DebugPrintf(const char* pszFormat, ...)
{
	va_list va;
	char psz[1000];
	va_start(va, pszFormat);
	int iLen = _vsnprintf(psz, 999, pszFormat, va);
	va_end(va);

	if (g_pfnMPrintf)
	{
		g_pfnMPrintf("%s\n", psz);
	}
	else
	{
		IDebugScrSrv* pDSS = static_cast<IDebugScrSrv*>(g_pScriptManager->GetService(IID_IDebugScriptService));
		if (!pDSS)
		{
			return;
		}
		cScrStr nil;
		pDSS->MPrint(psz, nil, nil, nil, nil, nil, nil, nil);
		pDSS->Release();
	}
#ifdef DEBUG
	ShowString(psz, 1000 + (iLen * 50));
#else
	iLen = iLen;
#endif
}

void DisplayPrintf(const char* pszFormat, ...)
{
	va_list va;
	char psz[1000];
	va_start(va, pszFormat);
	_vsnprintf(psz, 1000, pszFormat, va);
	va_end(va);

	ShowString(psz, CalcTextTime(psz, 500));
}


void MultiParmAssignInt(sMultiParm* pParm, int iValue)
{
	if (pParm->type == kMT_String || pParm->type == kMT_Vector)
	{
		g_pMalloc->Free(pParm->psz);
	}
	pParm->type = kMT_Int;
	pParm->i = iValue;
}

void MultiParmAssignFloat(sMultiParm* pParm, float fValue)
{
	if (pParm->type == kMT_String || pParm->type == kMT_Vector)
	{
		g_pMalloc->Free(pParm->psz);
	}
	pParm->type = kMT_Float;
	pParm->f = fValue;
}

void MultiParmAssignString(sMultiParm* pParm, const char* pszValue)
{
	if (pParm->type == kMT_String || pParm->type == kMT_Vector)
	{
		g_pMalloc->Free(pParm->psz);
		pParm->type = kMT_Undef;
	}
	pParm->psz = reinterpret_cast<char*>(g_pMalloc->Alloc(strlen(pszValue) + 1));
	if (pParm->psz)
	{
		strcpy(pParm->psz, pszValue);
		pParm->type = kMT_String;
	}
}

void MultiParmAssignVector(sMultiParm* pParm, const mxs_vector* pvValue)
{
	if (pParm->type == kMT_String || pParm->type == kMT_Vector)
	{
		g_pMalloc->Free(pParm->psz);
		pParm->type = kMT_Undef;
	}
	pParm->pVector = reinterpret_cast<mxs_vector*>(g_pMalloc->Alloc(sizeof(mxs_vector)));
	if (pParm->pVector)
	{
		memcpy(pParm->pVector, pvValue, sizeof(mxs_vector));
		pParm->type = kMT_Vector;
	}
}


