
#ifndef _SCRIPTVARS_H
#define _SCRIPTVARS_H

#include <cstring>
#include "lg/types.h"
#include "lg/scrmanagers.h"

extern IScriptMan* g_pScriptManager;


#define SCRIPT_VAR(className,varName) varName(#className, #varName)
#define SCRIPT_VAROBJ(className,varName, objId) varName(#className, #varName, objId)


class script_var
{
protected:
	bool			m_fDynName;
	sScrDatumTag	m_tag;

public:
	~script_var()
	{
		if (m_fDynName)
			delete[] m_tag.pszName;
	}
	script_var()
	{
		m_fDynName = false;
	}
	script_var(const char* pszScriptName, const char* pszVarName)
	{
		m_fDynName = false;
		m_tag.pszClass = pszScriptName;
		m_tag.pszName = pszVarName;
	}
	script_var(const char* pszScriptName, const char* pszVarName, int iObj)
	{
		m_fDynName = false;
		m_tag.pszClass = pszScriptName;
		m_tag.pszName = pszVarName;
		m_tag.objId = iObj;
	}

	void SetObj(int iObj)
	{
		m_tag.objId = iObj;
	}

	void SetDynTag(const char* pszDynName)
	{
		char* pszNew = new char[strlen(pszDynName)+1];
		if (pszNew)
		{
			strcpy(pszNew, pszDynName);
			if (m_fDynName)
				delete[] m_tag.pszName;
			m_tag.pszName = pszNew;
			m_fDynName = true;
		}
	}
	void Clear()
	{
		sMultiParm param;
		g_pScriptManager->ClearScriptData(&m_tag, &param);
	}
	script_var& operator= (const script_var& rCpy)
	{
		m_tag = rCpy.m_tag;
		return *this;
	}
};


class script_int : public script_var
{
public:
	script_int() : script_var()
	{ }
	script_int(const char* pszScriptName, const char* pszVarName)
		: script_var(pszScriptName,pszVarName)
	{ }
	script_int(const char* pszScriptName, const char* pszVarName, int iObj)
		: script_var(pszScriptName,pszVarName,iObj)
	{ }
	
	void Init(int iObj = 0)
	{
		if (!g_pScriptManager->IsScriptDataSet(&m_tag))
			*this = iObj;
	}

	operator int()
	{
		sMultiParm param;
		param.type = kMT_Int;
		g_pScriptManager->GetScriptData(&m_tag, &param);
		return param.i;
	}
	script_int& operator= (int iVal)
	{
		sMultiParm param;
		param.type = kMT_Int;
		param.i = iVal;
		g_pScriptManager->SetScriptData(&m_tag, &param);
		return *this;
	}
	script_int& operator+= (int iVal)
	{
		return *this = *this + iVal;
	}
	script_int& operator-= (int iVal)
	{
		return *this = *this - iVal;
	}
	script_int& operator++ ()
	{
		return *this = *this + 1;
	}
	script_int operator++ (int iPF)
	{
		script_int __tmp = *this;
		*this = *this + 1;
		return __tmp;
	}
	script_int& operator-- ()
	{
		return *this = *this - 1;
	}
	script_int operator-- (int iPF)
	{
		script_int __tmp = *this;
		*this = *this - 1;
		return __tmp;
	}
};


template<typename _Type>
class script_handle : public script_var
{
public:
	script_handle() : script_var()
	{ }
	script_handle(const char* pszScriptName, const char* pszVarName)
		: script_var(pszScriptName,pszVarName)
	{ }
	script_handle(const char* pszScriptName, const char* pszVarName, int iObj)
		: script_var(pszScriptName,pszVarName,iObj)
	{ }
	
	// gcc is being a snotty bitch over type casting.
	// so I figure that we want to initialize handles to NULL anyway
	void Init()
	{
		if (!g_pScriptManager->IsScriptDataSet(&m_tag))
		{
			sMultiParm param;
			param.type = kMT_Int;
			param.i = 0;
			g_pScriptManager->SetScriptData(&m_tag, &param);
		}
	}

	operator _Type()
	{
		sMultiParm param;
		param.type = kMT_Int;
		g_pScriptManager->GetScriptData(&m_tag, &param);
		return reinterpret_cast<_Type>(param.i);
	}
	script_handle<_Type>& operator= (_Type hVal)
	{
		sMultiParm param;
		param.type = kMT_Int;
		param.i = reinterpret_cast<int>(hVal);
		g_pScriptManager->SetScriptData(&m_tag, &param);
		return *this;
	}
};


class script_float : public script_var
{
public:
	script_float() : script_var()
	{ }
	script_float(const char* pszScriptName, const char* pszVarName)
		: script_var(pszScriptName,pszVarName)
	{ }
	script_float(const char* pszScriptName, const char* pszVarName, int iObj)
		: script_var(pszScriptName,pszVarName,iObj)
	{ }
	
	void Init(float fObj = 0)
	{
		if (!g_pScriptManager->IsScriptDataSet(&m_tag))
			*this = fObj;
	}

	operator float()
	{
		sMultiParm param;
		param.type = kMT_Float;
		g_pScriptManager->GetScriptData(&m_tag, &param);
		return param.f;
	}
	script_float& operator= (float fVal)
	{
		sMultiParm param;
		param.type = kMT_Float;
		param.f = fVal;
		g_pScriptManager->SetScriptData(&m_tag, &param);
		return *this;
	}
	script_float& operator+= (float fVal)
	{
		return *this = *this + fVal;
	}
	script_float& operator-= (float fVal)
	{
		return *this = *this - fVal;
	}
	script_float& operator++ ()
	{
		return *this = *this + 1;
	}
	script_float operator++ (int iPF)
	{
		script_float __tmp = *this;
		*this = *this + 1;
		return __tmp;
	}
	script_float& operator-- ()
	{
		return *this = *this - 1;
	}
	script_float operator-- (int iPF)
	{
		script_float __tmp = *this;
		*this = *this - 1;
		return __tmp;
	}
};


#endif // _SCRIPTVARS_H
