#include <fstream>
#include <cstring>
#include <windows.h>
#include <winver.h>

#include "lg/objstd.h"
#include "lg/interface.h"

#include "lg/types.h"
#include "lg/defs.h"
#include "lg/scrmsgs.h"
#include "lg/scrservices.h"
#include "lg/objects.h"
#include "lg/input.h"

#include "ScriptLib.h"
#include "ScriptModule.h"
#include "scrversion.h"

static char* GetScriptPaths(void);
static char* FindScriptModule(const char* pszName, const char* pszPaths);
static void DoSuccess(int iObjId);
static void DoFailure(int iObjId);

using namespace std;


const sScrClassDesc cScriptModule::sm_ScriptsArray[] = {
	{ "version", "VersionCheck", "CustomScript", cScr_VersionCheck::MakeVersionCheck },
};
const unsigned int cScriptModule::sm_ScriptsArraySize = 1;


IScript* __stdcall cScr_VersionCheck::MakeVersionCheck(char* pszName, int iHostObjId)
{
	cScr_VersionCheck* pscrRet = new(nothrow) cScr_VersionCheck(pszName, iHostObjId);
	if (pscrRet)
	{
		pscrRet->AddRef();
		pscrRet->InitScriptData();
	}
	return static_cast<IScript*>(pscrRet);
}

long __stdcall cScr_VersionCheck::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!static_cast<sSimMsg*>(pMsg)->fStarting)
		return 1;

	char* pszScriptPaths = GetScriptPaths();
	char* pszParams = GetObjectParams(m_iObjId);
	char* pszScript;
	char* pszToken = pszParams;
	for (pszScript = strsep(&pszToken, ";"); pszScript; pszScript = strsep(&pszToken, ";"))
	{
		if (!*pszScript)
			continue;
		DWORD scrVersHigh = 0, scrVersLow = 0;
		char* pszVers = strchr(pszScript, '=');
		if (pszVers)
		{
			char* pt = pszVers + 1;
			scrVersHigh = strtoul(pt, &pt, 10) << 16;
			if (pt && *pt == '.')
			{
				scrVersHigh |= strtoul(pt+1, &pt, 10);
				if (pt && *pt == '.')
				{
					scrVersLow = strtoul(pt+1, &pt, 10) << 16;
					if (pt && *pt == '.')
						scrVersLow |= strtoul(pt+1, &pt, 10);
				}
			}
			*pszVers = '\0';
		}
		char* pszScriptFile = FindScriptModule(pszScript, pszScriptPaths);
		if (!pszScriptFile)
		{
			DoFailure(m_iObjId);
			g_pMalloc->Free(pszParams);
			return 1;
		}
		if (scrVersHigh == 0 && scrVersLow == 0)
		{
			delete[] pszScriptFile;
			continue;
		}
		DWORD z;
		unsigned int len = ::GetFileVersionInfoSizeA(pszScriptFile, &z);
		if (len)
		{
			char* pBuffer;
			VS_FIXEDFILEINFO* pFileVers;
			pBuffer = new(nothrow) char[len];
			if (pBuffer)
			{
				::GetFileVersionInfoA(pszScriptFile, z, len, reinterpret_cast<void*>(pBuffer));
				len = 0;
				::VerQueryValueA(reinterpret_cast<void*>(pBuffer), "\\", reinterpret_cast<void**>(&pFileVers), &len);
				if (len > 0)
				{
					if ( (pFileVers->dwFileVersionMS < scrVersHigh)
					  || (pFileVers->dwFileVersionMS == scrVersHigh 
					   && pFileVers->dwFileVersionLS < scrVersLow)
					)
					{
						DoFailure(m_iObjId);
						delete[] pBuffer;
						delete[] pszScriptFile;
						g_pMalloc->Free(pszParams);
						if (pszScriptPaths)
							delete[] pszScriptPaths;
						return 1;
					}
				}
				delete[] pBuffer;
			}
		}
		else
		{
			DoFailure(m_iObjId);
			delete[] pszScriptFile;
			g_pMalloc->Free(pszParams);
			if (pszScriptPaths)
				delete[] pszScriptPaths;
			return 1;
		}
		delete[] pszScriptFile;
	}
	DoSuccess(m_iObjId);
	g_pMalloc->Free(pszParams);
	if (pszScriptPaths)
		delete[] pszScriptPaths;
	return 1;
}


// script_module_path
char* GetScriptPaths(void)
{
	char filepath[140];
	char line[80];		// Thief has this hardcoded. What a ridiculously arbitrary number.
	char* paths = NULL;
	//SInterface<IInputBinder> pIB(g_pScriptManager);
	//pIB->GetVarValue("script_module_path", buf, 80);
	//char* cmd = pIB->ProcessCmd("get script_module_path");
	//strcpy(buf, cmd);
	// It's never easy...
	filepath[0] = '\0';
	int cwdlen = ::GetCurrentDirectoryA(126, filepath);
	if (cwdlen > 0 && cwdlen < 126)
		strcat(filepath, "\\");
	// Ignore the path if it's really long, for whatever bizarre reason
	strcat(filepath, "darkinst.cfg");
	ifstream cfgfile(filepath);
	if (cfgfile.is_open())
	{
		while (cfgfile.getline(line,80))
		{
			if (!strnicmp(line,"script_module_path",18) && isspace(line[18]))
			{
				char* l = &line[19];
				while (isspace(*l)) ++l;
				char* k = l + strlen(l) - 1;
				while (isspace(*k)) --k;
				++k;
				paths = new(nothrow) char[k - l + 1];
				if (paths)
				{
					strncpy(paths,l,k - l);
					paths[k-l] = '\0';
				}
			}
		}
		cfgfile.close();
	}
	return paths;
}

char* FindScriptModule(const char* pszName, const char* pszPaths)
{
	if (!pszPaths || !*pszPaths)
	{
		if (::GetFileAttributesA(pszName) != 0xFFFFFFFF);
		{
			int len = strlen(pszName) + 1;
			char* ret;
			if (!strchr(pszName, '.'))
			{
				ret = new(nothrow) char[len + 4];
				if (ret)
				{
					strcpy(ret, pszName);
					strcat(ret, ".osm");
				}
			}
			else
			{
				ret = new(nothrow) char[len];
				if (ret)
					strcpy(ret, pszName);
			}
			return ret;
		}
		return NULL;
	}
	char filepath[96]; // length of buf + 8 + 3 + a little extra
	char *scrpaths = new char[strlen(pszPaths)+1];
	strcpy(scrpaths,pszPaths);
	char* tok = scrpaths;
	for (char* path = strsep(&tok, "+"); path; path = strsep(&tok, "+"))
	{
		while (isspace(*path)) ++path;
		if (!*path)
			continue;
		strcpy(filepath, path);
		strcat(filepath, "\\");
		strcat(filepath, pszName);
		if (!strchr(pszName, '.'))
			strcat(filepath, ".osm");
		if (::GetFileAttributesA(filepath) != 0xFFFFFFFF)
		{
			delete[] scrpaths;
			char* ret = new(nothrow) char[strlen(filepath)+1];
			if (ret)
				strcpy(ret,filepath);
			return ret;
		}
	}
	delete[] scrpaths;
	return NULL;
}

void DoSuccess(int iObjId)
{
	CDSend("TurnOn", iObjId);
	SInterface<IObjectSystem> pOS(g_pScriptManager);
	pOS->Destroy(iObjId);
}

void DoFailure(int iObjId)
{
	CDSend("TurnOff", iObjId);
	ShowBook(iObjId, true, 0x7FFFFFFF);
	SInterface<IObjectSystem> pOS(g_pScriptManager);
	pOS->Destroy(iObjId);
}

